package com.bolthole;

/**
 * This class is currently a wrapper around the
 * jlayout class, to turn it into a "real" layout manager.
 * It's primary function is to move things around.
 * It does not change a component's size, unless it is zero size.
 *
 * It give you the ability to either absolute-position something, or
 * specify its XY and/or updown coordinates relative to another object.
 * Types of relation include:
 *	TileLeftOf
 *	AlignLeft
 *	AlignInsideLeft
 *	Center
 *	TileRightOf
 *	AlignRight
 *	AlignInsideRight
 *
 * YOU MUST USE ONE OF THE FOLLOWING TWO EXAMPLES to use this
 * LayoutManager. Any other type of code structure will cause your objects
 * to not be mapped on the screen.
 *
 * This assumes you have a parent Panel "panel" class, and a
 * Component "obj" you want put in it. (and "obj2" 
 *
 * Example 1:
 *	flexilayout layout=new flexilayout();
 *	panel=new Panel(layout);
 *	panel.add(obj);
 *	layout.setLocation(obj,x,y);
 *
 * Example 2:
 *
 *	flexilayout layout=new flexilayout();
 *	panel=new Panel(layout);
 *	panel.add(obj);
 *	layout.setConstraints(obj,layout.TileFromRightOf,obj2,
 *				  layout.TileFromBottomOf,obj2);
 *
 * It is only acceptible to leave one of the Constraints pairs empty if
 * you have already set a position in that coordinate.
 *
 * Example of mixed positioning:
 *
 *	flexilayout layout=new flexilayout();
 *	panel=new Panel(layout);
 *	panel.add(obj);
 *	layout.setLocation(obj, 0, 100);
 *	layout.setConstraints(obj,layout.TileFromRightOf,obj2,
 *				  0,null);
 *
 * The above example will have the X position vary relative to obj2's position.
 * But it will always have a Y pixel address of 100.
 *
 *
 * LIMITATIONS: This layout manager is geared towards preserving sizes of
 * the managed components. So they will be completely unresponsive to
 * the size of their container.
 * On the brigher size, you CAN have the container resize itself to match
 * the size of the components inside it, by Frame.pack(), or calling
 * flexilayout.minimumLayoutSize() directly.
 */
 

import java.awt.*;

public class flexilayout extends jlayout implements LayoutManager {

	public void setConstraints(Component targ,
			      int xystyle, Object xyobj,
			      int updownstyle, Object updownobj) {

		if(getobjectindex(targ)==-1){
			addObject(targ);
		}
		super.setConstraints(targ,xystyle,xyobj,updownstyle,updownobj);
	}

	public void setLocation(Component targ, int x, int y){
		if(getobjectindex(targ)==-1){
			addObject(targ);
		}
		super.setLocation(targ,x,y);
	}

	/**************************************************/
	/* Old fuddy-duddy LayoutManager stuff */
	/**************************************************/

	public void addLayoutComponent(String name, Component com){
		System.out.println("flexilayout Supposedly adding object");
		addObject(com);
	}

	/* This gets called the first time a component is displayed,and
	 * every time a component wants to be resized.
	 *
	 * According to javasoft docs, I should use
	 * move(), resize(), and reshape() here.
	 * [deprecated: newstuff= setLocation(), setSize(), setBounds()]
	 *
	 * But apparently, the damn thing wont SHOW UP unless we
	 * call setBounds() explicitly. 
	 */
	public void layoutContainer(Container con){
		int objmax;
		Point newloc;
		Component component;
		Dimension tmpsize;

	System.out.println("DOING layoutContainer()");

	    synchronized(objectlist){
		objmax=objectlist.size();

		/* Damn, this is annoying. frigging @#$@$@!!! */
		/* It seems that calling Container.add(Component) does
		 * not call ANY LayoutManager routine. So we have
		 * to iterate through all the Components in the
		 * Container in case we havent added them yet. ARRRGGG!!
		 */

		/* Unfortunately, we do not know WHICH ONE, if any,
		 * of the contained components may have resized itself.
		 * So, have to update the size cache for ALL of them
		 */
		for(int objcount=0;objcount<objmax;objcount++){
			component=(Component)getobject(objcount);
			tmpsize=component.getSize();
			if((tmpsize.width==0)||
			   (tmpsize.height==0)) {
			   	System.out.println("using getMINIMUM on "+component);
				tmpsize=component.getMinimumSize();
			   	System.out.println("   "+tmpsize);
			}
			/* This just sets the CACHED size. It doesnt change
			 * the object size
			 */
			setsize(objcount, tmpsize);
		}
	    }
		layoutAll();

	    synchronized(objectlist){

		/* Now go move them around in the visible world. */
		for(int objcount=0;objcount<objmax;objcount++){
			component=(Component)getobject(objcount);
			newloc=getpoint(objcount);
			tmpsize=getsize(objcount);
			if((newloc==component.getLocation()) &
			   (tmpsize==component.getSize())){
			   	continue;
			}

			component.setBounds(newloc.x,newloc.y,
						tmpsize.width,tmpsize.height);

		}
	    }
				
	}

	public Dimension minimumLayoutSize(Container con){
		return getBoundingDimension();
		
	}
	public Dimension preferredLayoutSize(Container con){
		return getBoundingDimension();
	}

	public void removeLayoutComponent(Component com){
		removeObject(com);
	}

	/*************************************************************
	 * New LayoutManager2 stuff. That I think is completely
	 * useless, and imposes more burden, so I may as well NOT
	 * implement it
	 *
	 *************************************************************/

/********
	public void addLayoutComponent(Component con,Object obj){
	}

	public float getLayoutAlignmentX(Container con){
	}
	float getLayoutAlignmentY(Container con){
	}

	void invalidateLayout(Container con){
	}

	public Dimension maximumLayoutSize(Container con){
	}

**************/

	
}
