package com.bolthole;

/**
 * This class is NOT actually a java.awt.LayoutManager compliant
 * class. This is a more free-form thing that I plan to use in a
 * word-processor. BUT... it can very easily be adapted for
 * LayoutManager type things. See the sample that comes along
 * with it, "lcontainer".
 *
 * I'll probably make a LayoutManager wrapper for it soon
 * Philip Brown
 * phil@bolthole.com Feb 2001
 *
 */

/* Internals comments about jlayhout:
 * Internally, we current keep "index positions" for each object.
 * So, thigns that deal with a specific Object, generaly do not need
 * mutexes around them. Even after you delete a particular object,
 * the others currently stay in the same position.
 * Just beware if you delete an object, then ADD an object immediately after
 * that, and at the same time you are doing a relayout or something.
 * There may be a race condition hit, where you would position the new
 * object based on the old object's info or something.
 *
 */

import java.util.*;
import java.awt.Point;
import java.awt.Dimension;



/**
 * Meaning of constraints;
 *  XYStyle goes with XYRef.
 *  UpDownStyle goes with UpDownRef.
 *  For XY coords, align the object bound with this constaints instance,
 *  in the manner indicated by XYStyle.
 *  (See the list of styles in jlayout)
 *  Same principle for UpDownStyle/UpDownRef
 *
 *  If Style value ==0, object is not consulted.
 *  If Style value !=0, object had better not be null.
 */
class constraints {
	int XYStyle=0, UpDownStyle=0;
	public Object XYRef, UpDownRef;
}


public class jlayout {

	/* "Style" values to use with private "constraints" class */
	/* for "Up", substitute "Left" */
	/* for "Down", substitute "Right" */
	public static final int TileLeftOf=1;
	public static final int TileUpOf=1;
	public static final int AlignLeft=2;
	public static final int AlignUp=2;
	public static final int AlignInsideLeft=3; /* use spacing */
	public static final int AlignInsideUp=3; /* use spacing */
	public static final int Center=10;
	public static final int TileRightOf=11;
	public static final int TileDownOf=11;
	public static final int AlignRight=12;
	public static final int AlignDown=12;
	public static final int AlignInsideRight=13; /* use spacing */
	public static final int AlignInsideDown=13; /* use spacing */


	Vector objectlist; /* This is the index master for all the
			    * other 'lists'
			    */
	Vector locationlist;
	Vector sizelist;
	Vector constraintslist;

	int border=5; /* Global "border" parameter, aka "spacing" */

	public jlayout(){
		objectlist=new Vector();
		locationlist=new Vector();
		sizelist=new Vector();
		constraintslist=new Vector();
	}

	boolean debug_flag=false;
	void debug(String msg){
		if(debug_flag==false)
			return;
		System.out.println(msg);
	}

	
	/**
	 * If object is already managed, will do nothing.
	 * If object is null, will do nothing.
	 * You MUST at minimum call this.setSize(obj,w,h) before
	 * doing another layout
	 */
	public void addObject(Object obj){
		addObject(obj,0,0,0,0);
	}

	/** long form of addObject()
	 * It is semi-acceptible to have x and y as 0.
	 * But you really, really should specify width and height here.
	 * We dont know what an object's size is unless you explicitly
	 * tell us. It's rather difficult to layout a 1.dimentional object.
	 *
	 * Tricks:
	 *   1. if you are arranging objects on a "page", add the first
	 *      object as a fake frame for the entire page. That's the
	 *      only way you can have relations to "page" edges
	 *
	 *   2. If you have anchor objects, make sure they are added "FIRST"
	 *      Objects are re-organized in order. The order does not change.
	 *      An object added early, will always be at or near the first
	 *      object to be re-laid-out.
	 *      
	 */
	public void addObject(Object obj, int x, int y, int width, int height){
		if(obj==null) return;
		if(objectlist.contains(obj)){
			return;
		}
		/* XXX NEED LOCKING HERE */
		objectlist.addElement(obj);
		locationlist.addElement(new Point(x,y));
		sizelist.addElement(new Dimension(width,height));
		constraintslist.addElement(null);
		debug("jlayout: added object("+getsize(locationlist.size()-1) +")");
		
	}


	/* At one point, I was planning to just set the element to null.
	 * But I dont see any motivation for that any more.
	 * If I ever migrated to having constraints by INDEX instead of
	 * by Object reference, this would make sense. But
	 * I dont do that right now.
	 */
	public void removeObject(Object obj){
		synchronized(objectlist){
			int pos=objectlist.indexOf(obj);
			if(pos==-1) return;
			objectlist.removeElementAt(pos);
			locationlist.removeElementAt(pos);
			sizelist.removeElementAt(pos);
			constraintslist.removeElementAt(pos);
		}
	}
	
	/** Note that we pass back a DIRECT REFERENCE to the
	 * internal Location object we use. This is for speed
	 * purposes.
	 * Dont mess with it!!!
	 * Messing with it will only screw things up for you
	 */
	public Point getLocation(Object obj){
		int pos=objectlist.indexOf(obj);
		if(pos==-1) return null;
		return (Point)locationlist.elementAt(pos);
	}
	/** Note that we pass back a DIRECT REFERENCE to the
	 * internal Location object we use. This is for speed
	 * purposes.
	 * Dont mess with it!!!
	 * Messing with it will only screw things up for you
	 */
	public Dimension getDimension(Object obj){
		int pos=objectlist.indexOf(obj);
		if(pos==-1) return null;
		return (Dimension)sizelist.elementAt(pos);
	}

	/**
	 * given a target object that must already be added into us,
	 * set constraints for it.
	 *
	 * You can currently set constraints of being from left, right,
	 * bottom, or top of an object.
	 * For any given direction, you can either have a single object
	 * to be that direction of, or pass null for that direction constraint.
	 *
	 * You SHOULD specify only objects that are managed by this
	 * same layout instance. Otherwise, they will be ignored as constraints.
	 */
	public void setConstraints(Object targ,
			      int xystyle, Object xyobj,
			      int updownstyle, Object updownobj) {

		constraints cons;
		int pos=objectlist.indexOf(targ);
		if(pos==-1){
			debug("Warning: attempt to set constraints for object we dont manage");
			return;
		}
		if(pos >= constraintslist.size()) return;
		cons=(constraints)constraintslist.elementAt(pos);
		if(cons==null) cons=new constraints();
		cons.XYStyle=xystyle;
		cons.XYRef=xyobj;
		cons.UpDownStyle=updownstyle;
		cons.UpDownRef=updownobj;
		constraintslist.setElementAt(cons,pos);
		
	}
			      

	/**
	 * It's up to the caller to call layoutAll() after this.
	 */
	public void setSize(Object targ, int width, int height){
		Dimension targsize;
		int pos=objectlist.indexOf(targ);
		if(pos==-1){
			debug("WARNING: ignoring setSize() for object we dont manage");
			return;
		}
		if(pos>=sizelist.size()) return;
		targsize=getsize(pos);
		targsize.width=width;
		targsize.height=height;
	
	}

	/**
	 * Do not call this on anything that has any Constraints set on
	 * it. You should only call this to absolute-position something as
	 * a special "anchor". Then most everything else should be
	 *
	 * It's up to the caller to call layoutAll() after this.
	 */
	public void setLocation(Object targ, int x, int y){
		Point targloc;
		int pos=objectlist.indexOf(targ);
		if(pos==-1) return;
		if(pos>=locationlist.size()) return;
		if(debug_flag==true){
			Object consobj=constraintslist.elementAt(pos);
			if(consobj!=null){
				debug("WARNING: ignoring setLocation on object with constraints");
			}
		}
		targloc=getpoint(pos);
		targloc.x=x;
		targloc.y=y;
	
	}


	/** This sets the GLOBAL "spacing" property.
	 *  This is only internal spacing between components, not external.
	 *  It is the caller's responsability to call layoutAll() after
	 *  setting this value
	 */
	public void setSpacing(int spacing){
		border=spacing;
	}


	/* A concession to make it easier for a true java.awt.LayoutManager
	 * wrapper
	 * Just go through and find the most extreme X and Y coords.
	 */
	public Dimension getBoundingDimension(){
		Dimension tmpsize;
		Point tmploc;
		int maxx=0,maxy=0;
		int tmpx,tmpy;
		synchronized(objectlist){
			int numobjs=objectlist.size();
			for(int objcount=0;objcount<numobjs; objcount++){				
				tmpsize=getsize(objcount);
				if(tmpsize==null) continue;
				tmploc=getpoint(objcount);
				if(tmploc==null) continue;
				tmpx=tmploc.x+tmpsize.width;
				tmpy=tmploc.y+tmpsize.height;
				if(tmpx>maxx) maxx=tmpx;
				if(tmpy>maxy) maxy=tmpy;
			}
		}
		return new Dimension(maxx,maxy);
	}

	/** layoutAll goes through all the objects we know about,
	 * and works out positions for all of them.
	 * You then need to go call this.getLocation() and this.getDimension()
	 * for each of them to find out new locations, if you desire.
	 *
	 * (Maybe I should make it optional to set their locations.
	 *  possibly do a if(instanceof something))
	 *
	 */
	public void layoutAll(){
		synchronized(objectlist){
			sweepObjs();
		}
	}

/**********************************************************************/
/**********************************************************************/
/*          PRIVATE stuff down here                                   */
/**********************************************************************/
/**********************************************************************/
	
	/* helper for LayoutManager wrapper only */
	void setsize(int pos, Dimension size){
		sizelist.setElementAt(size,pos);
	}
	/* helper for LayoutManager wrapper only */
	Object getobject(int pos){
		return objectlist.elementAt(pos);
	}
	/* helper for LayoutManager wrapper only */
	int getobjectindex(Object obj){
		return objectlist.indexOf(obj);
	}


	/* internal casting routine */
	Point getpoint(int pos){
		return (Point)locationlist.elementAt(pos);
	}
	/* internal casting routine */
	Dimension getsize(int pos){
		return (Dimension)sizelist.elementAt(pos);
	}


	/**
	 * subroutine of layoutAll();
	 * This kinda assumes the objects are already vaguely
	 * in sorted order. Otherwise things may be unhappy.
	 * sweeps through objects, and checks left/right ordering.
	 * Adjusts x position as appropriate... ONCE.
	 * If there are loops, too bad for you, we wont go infinite.
	 *
	 * Returns true if something moved, false if not
	 */
	boolean sweepObjs(){
		int objmax=objectlist.size();
		boolean changed=false;
		Object thisobj, refobj;
		constraints consobj;
		Point loc;
		Dimension size;
		for(int objcount=0; objcount<objmax; objcount++){
			thisobj=objectlist.elementAt(objcount);
			consobj=(constraints)constraintslist.elementAt(objcount);
			if(consobj ==null){
				debug(" Debug: absolute positioning only for "+thisobj);
				debug( "     "+getpoint(objcount)+"  "+
				      getsize(objcount));
				continue;
			}
			loc=getpoint(objcount);
			size=getsize(objcount);

			if(AlignXY(consobj, loc, size)==true){
				changed=true;
			}
			if(AlignUpDown(consobj,loc, size)==true){
				changed=true;
			}

			if(changed==true){
				debug("new coords set for object "+thisobj);
				debug("  new loc supposedly "+loc);
			}
		}
		return changed;
	
	}

	/** subroutine of sweepObjs()
	 * For a particular object, check if there's an object
	 * we should be RightOf
	 * If yes, adjust x parameter in locationlist, and return true
	 *   if different from our original pos.
	 * Otherwise, return false
	*/
	boolean AlignXY(constraints consobj, Point currloc, Dimension size){
		Object refobj;
		int refindex=-1;
		int mycentre,refcentre;
		int newx;

		refobj=consobj.XYRef;
		if(refobj==null){
			debug("ERROR: null obj for alignXY");
			return false;
		}
		refindex=objectlist.indexOf(refobj);

		if(refindex == -1){
			debug("ERROR: non-managed obj for alignXY: "+refobj);
			return false;
		}

		/* Start of main "switch" in this routine */
		switch(consobj.XYStyle){

		case 0:
			return false;
		
		case Center:
			mycentre=size.width >>1;
			refcentre=getpoint(refindex).x;
			refcentre+=getsize(refindex).width>>1;
			newx=refcentre -mycentre;
			break;

		case TileLeftOf:
			newx=getpoint(refindex).x;
			newx-= size.width +border;
			break;

		case  TileRightOf:
			newx=getpoint(refindex).x;
			newx+=getsize(refindex).width + border;
			break;

		case AlignLeft:
			newx=getpoint(refindex).x;
			break;

		case AlignRight:
			newx=getpoint(refindex).x;
			newx+=getsize(refindex).width;
			newx-=size.width;
			break;
		default:
			debug("ERROR: unknown case for AlignXY()");
			return false;
		}

		if(newx==currloc.x){
			return false;
		}
		currloc.x=newx;
		return true;
	
	}


	/** subroutine of sweepObjs()
	 * For a particular object, check if there's an object
	 * we should be RightOf
	 * If yes, adjust x parameter in locationlist, and return true
	 *   if different from our original pos.
	 * Otherwise, return false
	*/
	boolean AlignUpDown(constraints consobj, Point currloc, Dimension size){
		Object refobj;
		int refindex=-1;
		int mycentre,refcentre;
		int newy;

		refobj=consobj.UpDownRef;
		if(refobj==null){
			debug("ERROR: null obj for alignUpDown");
			return false;
		}
		refindex=objectlist.indexOf(refobj);

		if(refindex == -1){
			debug("ERROR: non-managed obj for alignUpDown: "+refobj);
			return false;
		}

		/* Start of main "switch" in this routine */
		switch(consobj.UpDownStyle){

		case 0:
			return false;
		
		case Center:
			mycentre=size.height >>1;
			refcentre=getpoint(refindex).y;
			refcentre+=getsize(refindex).height>>1;
			newy=refcentre -mycentre;
			break;

		case TileLeftOf:
			newy=getpoint(refindex).y;
			newy-= size.height +border;
			break;

		case  TileRightOf:
			newy=getpoint(refindex).y;
			newy+=getsize(refindex).height + border;
			break;

		case AlignLeft:
			newy=getpoint(refindex).y;
			break;

		case AlignRight:
			newy=getpoint(refindex).y;
			newy+=getsize(refindex).height;
			newy-=size.height;
			break;
		default:
			debug("ERROR: unknown case for AlignUpDown()");
			return false;
		}

		if(newy==currloc.y){
			return false;
		}
		currloc.y=newy;
		return true;
	
	}

}
